/*
//@HEADER
// ************************************************************************
//
// logger.hpp
//                     		  Pressio
//                         Copyright 2019
//    National Technology & Engineering Solutions of Sandia, LLC (NTESS)
//
// Under the terms of Contract DE-NA0003525 with NTESS, the
// U.S. Government retains certain rights in this software.
//
// Pressio is licensed under BSD-3-Clause terms of use:
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
//
// 1. Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
//
// 2. Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
//
// 3. Neither the name of the copyright holder nor the names of its
// contributors may be used to endorse or promote products derived
// from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
// FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
// COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
// INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
// SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
// HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
// STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
// IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.
//
// Questions? Contact Francesco Rizzi (fnrizzi@sandia.gov)
//
// ************************************************************************
//@HEADER
*/

#ifndef PRESSIO4PY_PYBINDINGS_LOGGER_HPP_
#define PRESSIO4PY_PYBINDINGS_LOGGER_HPP_

namespace pressio4py{

void bindLogger(pybind11::module & mParent)
{
  // make sure to redirect stdout/stderr streams to python stdout
  pybind11::add_ostream_redirect(mParent, "ostream_redirect");

  // create logger submodule
  pybind11::module loggerModule = mParent.def_submodule("logger");
  pybind11::enum_<pressio::logto>(loggerModule, "logto")
    .value("terminal", pressio::logto::terminal)
    .export_values();

  pybind11::enum_<pressio::log::level>(loggerModule, "loglevel")
    .value("trace",	pressio::log::level::trace)
    .value("debug",	pressio::log::level::debug)
    .value("info",	pressio::log::level::info)
    .value("warn",	pressio::log::level::warn)
    .value("err",	pressio::log::level::err)
    .value("critical",	pressio::log::level::critical)
    .value("off",	pressio::log::level::off)
    .export_values();

  loggerModule.def("setVerbosity",
		   &pressio::log::setVerbosity<std::vector<::pressio::log::level>>);
  loggerModule.def("initialize",
   		   &pressio::log::initialize);
  loggerModule.def("initialize",
   		   &pressio::log::initialize2);
  loggerModule.def("finalize",
   		   &pressio::log::finalize<void>);
}

} //end namespace pressio4py
#endif
